#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <libubox/list.h>
#include <libubox/uloop.h>
#include <libubus.h>
#include <uci.h>
#include <sched.h>

#include <easy/easy.h>

#include "debug.h"
#include "host_config.h"
#include "topologyd.h"
#include "host.h"
#include "mdns_avahi.h"
#include "host_nodes.h"
#include <libnetfilter_conntrack/libnetfilter_conntrack.h>
#include "host_wan_stats.h"

void host_node_destroy(struct host_node *n)
{
	free(n);
}

struct host_node *host_node_create(uint8_t *hwaddr)
{
	struct host_node *n;

	n = calloc(1, sizeof(struct host_node));
	if (!n) {
		warn("OOM: Node alloc failed!\n");
		return NULL;
	}

	memcpy(n->hwaddr, hwaddr, 6);
	memcpy(n->link_hwaddr, hwaddr, 6); /* set by topology discovery */

	return n;
}

struct host_node *host_node_lookup(struct hlist_head *table, uint8_t *hwaddr)
{
	int hidx = node_hash(hwaddr);
	struct host_node *n = NULL;

	if (hwaddr_is_zero(hwaddr))
		return NULL;

	hlist_for_each_entry(n, &table[hidx], hlist) {
		if (!memcmp(hwaddr, n->hwaddr, 6))
			return n;
	}

	return NULL;
}

int host_node_del(struct hlist_head *table, uint8_t *hwaddr, struct topologyd_private *priv)
{
	int hidx = node_hash(hwaddr);
	struct host_node *n;

	n = host_node_lookup(table, hwaddr);
	if (!n)
		return -1;

	hlist_del(&n->hlist, &table[hidx]);
	dbg("Node " MACFMT " removed from topology\n", MAC2STR(n->hwaddr));

	host_node_destroy(n);
	return 0;
}

struct host_node *host_node_add(struct hlist_head *table, uint8_t *hwaddr)
{
	struct host_node *n;

	if (hwaddr_is_mcast(hwaddr)) {
		dbg("Node " MACFMT" is a bcast or mcast address\n", MAC2STR(hwaddr));
		return NULL;
	}


	n = host_node_lookup(table, hwaddr);
	if (n) {
		dbg("Node " MACFMT " already in topology\n", MAC2STR(hwaddr));
		return n;
	}

	n = host_node_create(hwaddr);
	if (n) {
		int hidx = node_hash(hwaddr);

		hlist_add_head(&n->hlist, &table[hidx]);
		dbg("Node " MACFMT " added to topology\n", MAC2STR(hwaddr));
	}

	return n;
}

void host_node_print_all(struct hlist_head *table)
{
	struct host_node *n;
	int i;

	dbg("Nodes:\n");
	for (i = 0; i < NODE_HTABLE_SIZE; i++) {
		if (hlist_empty(&table[i]))
			continue;

		hlist_for_each_entry(n, &table[i], hlist) {
			dbg("Node: " MACFMT "\n", MAC2STR(n->hwaddr));
		}
	}
}

static void host_node_ct_node_destroy(struct host_nf_tc *nf_tc)
{
	free(nf_tc);
}

static void host_node_nfct_copy(struct host_nf_tc *host_ct, struct nf_conntrack *ct)
{
	host_ct->id = nfct_get_attr_u32(ct, ATTR_ID);
	host_ct->start_timestamp = nfct_get_attr_u64(ct, ATTR_TIMESTAMP_START);
}

static struct host_nf_tc *host_node_ct_node_create(struct nf_conntrack *ct)
{
	struct host_nf_tc *n = NULL;

	//allocated memory free is done in function host_node_clean_nf_tc
	n = calloc(1, sizeof(struct host_nf_tc));
	if (!n) {
		warn("OOM: Node alloc failed!\n");
		return NULL;
	}

	host_node_nfct_copy(n, ct);

	return n;
}

static struct host_nf_tc_ip_list *host_node_ct_ip_node_create(struct nf_conntrack *ct)
{
	struct host_nf_tc_ip_list *n_ip;

	//aalocated memory free is done in function host_node_clean_nf_tc
	n_ip = calloc(1, sizeof(struct host_nf_tc_ip_list));
	if (!n_ip) {
		warn("OOM: Node alloc failed!\n");
		return NULL;
	}
	return n_ip;
}

static struct host_nf_tc *host_node_ct_node_lookup(struct topologyd_private *priv, struct nf_conntrack *ct)
{
	struct host_nf_tc *n = NULL;
	unsigned int id;
	unsigned long long start_timestamp;

	id = nfct_get_attr_u32(ct, ATTR_ID);
	start_timestamp = nfct_get_attr_u64(ct, ATTR_TIMESTAMP_START);
	hlist_for_each_entry(n, &priv->host.ct_node_htable[ct_node_hash(id)], hlist) {
		if (id == n->id && start_timestamp == n->start_timestamp)
			return n;
	}

	return NULL;
}

static struct host_nf_tc_ip_list *host_node_ct_ip_node_lookup(struct topologyd_private *priv, struct nf_conntrack *ct)
{
	struct host_nf_tc_ip_list *n = NULL;
	struct in_addr addr;
	struct nfct_attr_grp_ipv4 grp_ipv4;
	int ret_val;

	ret_val = nfct_get_attr_grp(ct, ATTR_GRP_ORIG_IPV4, &grp_ipv4);
	if (ret_val || priv == NULL)
		return NULL;

	memcpy(&addr, &grp_ipv4.src, sizeof(grp_ipv4.src));
	addr.s_addr = ntohl(addr.s_addr);
	hlist_for_each_entry(n, &priv->host.ct_ip_node_htable[ct_ip_node_hash(addr.s_addr)], hlist) {
		if (addr.s_addr == n->ipv4_addr.s_addr)
			return n;
	}

	return NULL;
}

static void host_get_ipv6_int_to_char_array(const unsigned int addr1[], char addr2[])
{
	int  addr_part;

	addr_part = (addr1[0]);
	memcpy(&addr2[0], &addr_part, sizeof(int));
	addr_part = (addr1[1]);
	memcpy(&addr2[4], &addr_part, sizeof(int));
	addr_part = (addr1[2]);
	memcpy(&addr2[8], &addr_part, sizeof(int));
	addr_part = (addr1[3]);
	memcpy(&addr2[12], &addr_part, sizeof(int));
}

static struct host_nf_tc_ip_list *host_node_ct_ip6_node_lookup(struct topologyd_private *priv, struct nf_conntrack *ct)
{
	struct host_nf_tc_ip_list *n = NULL;
	struct nfct_attr_grp_ipv6 grp_ipv6;
	char ipv6_addr[sizeof(int)*4];
	int ret_val;

	ret_val = nfct_get_attr_grp(ct, ATTR_GRP_ORIG_IPV6, &grp_ipv6);
	if (ret_val || priv == NULL)
		return NULL;
	host_get_ipv6_int_to_char_array(grp_ipv6.src, ipv6_addr);

	hlist_for_each_entry(n, &priv->host.ct_ipv6_node_htable[ct_ip_node_hash(grp_ipv6.src[3])], hlist) {
		if (!memcmp(ipv6_addr, n->ipv6_addr.sin6_addr.s6_addr, sizeof(int)*4))
		{
			return n;
		}
	}

	return NULL;
}

static struct host_nf_tc *host_node_ct_node_add(struct nf_conntrack *ct, struct topologyd_private *priv)
{
	struct host_nf_tc *n = NULL;

	n = host_node_ct_node_create(ct);
	if (!n)
		return NULL;
	hlist_add_head(&n->hlist, &priv->host.ct_node_htable[ct_node_hash(n->id)]);
	n->in_use = 1;

	return n;
}


static struct host_nf_tc_ip_list *host_node_ct_ip_node_add(struct nf_conntrack *ct, struct topologyd_private *priv)
{
	struct host_nf_tc_ip_list *n_ip = NULL;
	struct nfct_attr_grp_ipv4 grp_ipv4;
	int ret_val;

	ret_val = nfct_get_attr_grp(ct, ATTR_GRP_ORIG_IPV4, &grp_ipv4);
	if (ret_val || priv == NULL)
		return NULL;

	n_ip = host_node_ct_ip_node_create(ct);
	if (!n_ip)
		return NULL;
	memcpy(&n_ip->ipv4_addr, &grp_ipv4.src, sizeof(grp_ipv4.src));
	n_ip->ipv4_addr.s_addr = ntohl(n_ip->ipv4_addr.s_addr);
	hlist_add_head(&n_ip->hlist, &priv->host.ct_ip_node_htable[ct_ip_node_hash(n_ip->ipv4_addr.s_addr)]);
	n_ip->in_use = HOST_WAN_STATS_IP_NODE_INIT;
	return n_ip;
}

static struct host_nf_tc_ip_list *host_node_ct_ipv6_node_add(struct nf_conntrack *ct, struct topologyd_private *priv)
{
	struct host_nf_tc_ip_list *n_ip = NULL;
	struct nfct_attr_grp_ipv6 grp_ipv6;
	char ipv6_src_addr[sizeof(int)*4];
	int ret_val;

	ret_val = nfct_get_attr_grp(ct, ATTR_GRP_ORIG_IPV6, &grp_ipv6);
	if (ret_val || priv == NULL)
		return NULL;

	host_get_ipv6_int_to_char_array(grp_ipv6.src, ipv6_src_addr);
	n_ip = host_node_ct_ip_node_create(ct);
	if (!n_ip)
		return NULL;

	memcpy(&n_ip->ipv6_addr, ipv6_src_addr, sizeof(ipv6_src_addr));
	hlist_add_head(&n_ip->hlist, &priv->host.ct_ipv6_node_htable[ct_ip_node_hash(grp_ipv6.src[3])]);
	n_ip->in_use = 1;
	return n_ip;
}

static int host_node_update_stats(struct nf_conntrack *ct, struct host_nf_tc *ct_node, struct host_nf_tc_ip_list *ct_ip_node)
{
	struct nfct_attr_grp_ctrs grp_ctrs_tx;
	struct nfct_attr_grp_ctrs grp_ctrs_rx;

	memset(&grp_ctrs_tx, 0, sizeof(struct nfct_attr_grp_ctrs));
	memset(&grp_ctrs_rx, 0, sizeof(struct nfct_attr_grp_ctrs));

	nfct_get_attr_grp(ct, ATTR_GRP_ORIG_COUNTERS, &grp_ctrs_tx);
	nfct_get_attr_grp(ct, ATTR_GRP_REPL_COUNTERS, &grp_ctrs_rx);

	if (!ct_node)
		return NFCT_CB_CONTINUE;

	ct_node->in_use = 1;

	if (!ct_ip_node)
		return NFCT_CB_CONTINUE;

	if (grp_ctrs_tx.packets < ct_node->tx_packets || grp_ctrs_rx.packets < ct_node->rx_packets) {
		ct_ip_node->tx_packets += grp_ctrs_tx.packets;
		ct_ip_node->tx_bytes += grp_ctrs_tx.bytes;
		ct_ip_node->rx_packets += grp_ctrs_rx.packets;
		ct_ip_node->rx_bytes += grp_ctrs_rx.bytes;
	} else {
		ct_ip_node->tx_packets += (grp_ctrs_tx.packets - ct_node->tx_packets);
		ct_ip_node->tx_bytes += (grp_ctrs_tx.bytes - ct_node->tx_bytes);
		ct_ip_node->rx_packets += (grp_ctrs_rx.packets - ct_node->rx_packets);
		ct_ip_node->rx_bytes += (grp_ctrs_rx.bytes - ct_node->rx_bytes);
	}
	ct_node->tx_packets = grp_ctrs_tx.packets;
	ct_node->rx_packets = grp_ctrs_rx.packets;
	ct_node->tx_bytes = grp_ctrs_tx.bytes;
	ct_node->rx_bytes = grp_ctrs_rx.bytes;

	ct_ip_node->in_use = 1;
	ct_ip_node->ct_running_count++;

        return NFCT_CB_CONTINUE;
}

static int host_node_ipv6_stats_received(enum nf_conntrack_msg_type type,
					struct nf_conntrack *ct, void *data)
{
	struct topologyd_private *priv = (struct topologyd_private *)data;
	struct nfct_attr_grp_ipv6 grp_ipv6;
	char ipv6_dst_addr[sizeof(int)*4];
	struct device_address_list *n;
	struct host_nf_tc *ct_node = NULL;
	struct host_nf_tc_ip_list *ct_ip_node = NULL;
	int ret_val;

	ret_val = nfct_get_attr_grp(ct, ATTR_GRP_ORIG_IPV6, &grp_ipv6);
	if (ret_val)
		return NFCT_CB_CONTINUE;

	host_get_ipv6_int_to_char_array(grp_ipv6.dst, ipv6_dst_addr);

	if (!hlist_empty(&priv->device_addr_list_table)) {
		hlist_for_each_entry(n, &priv->device_addr_list_table, hlist) {
			/* cppcheck-suppress uninitvar */
			if (n->addr.family != AF_INET6)
				continue;
			if (!memcmp(ipv6_dst_addr, n->addr.ipv6_sin_addr.s6_addr, sizeof(ipv6_dst_addr)))
                		return NFCT_CB_CONTINUE;
		}
	}

	ct_node = host_node_ct_node_lookup(priv, ct);
	if (ct_node) {
		ct_ip_node = host_node_ct_ip6_node_lookup(priv, ct);
		if(!ct_ip_node)
			ct_ip_node = host_node_ct_ipv6_node_add(ct, priv);
		return host_node_update_stats(ct, ct_node, ct_ip_node);
	} else {
		ct_node = host_node_ct_node_add(ct, priv);
		ct_ip_node = host_node_ct_ipv6_node_add(ct, priv);
		return host_node_update_stats(ct, ct_node, ct_ip_node);
	}
}

static int host_node_stats_received(enum nf_conntrack_msg_type type,
					struct nf_conntrack *ct, void *data)
{
	struct topologyd_private *priv = (struct topologyd_private *)data;
	struct host_nf_tc *ct_node = NULL;
	struct host_nf_tc_ip_list *ct_ip_node = NULL;
	struct device_address_list *n = NULL;
	struct nfct_attr_grp_ipv4 grp_ipv4;

	nfct_get_attr_grp(ct, ATTR_GRP_ORIG_IPV4, &grp_ipv4);

	if (!hlist_empty(&priv->device_addr_list_table)) {
		hlist_for_each_entry(n, &priv->device_addr_list_table, hlist) {
			if (n->addr.family != AF_INET)
				continue;
			if (ntohl(grp_ipv4.dst) == n->addr.ipv4_sin_addr.s_addr)
                		return NFCT_CB_CONTINUE;
		}
	}
	ct_node = host_node_ct_node_lookup(priv, ct);
	if (ct_node) {
		ct_ip_node = host_node_ct_ip_node_lookup(priv, ct);
		if(!ct_ip_node)
			ct_ip_node = host_node_ct_ip_node_add(ct, priv);
		return host_node_update_stats(ct, ct_node, ct_ip_node);
	} else {
		ct_node = host_node_ct_node_add(ct, priv);
		ct_ip_node = host_node_ct_ip_node_add(ct, priv);
		return host_node_update_stats(ct, ct_node, ct_ip_node);
	}
}

static void host_node_nfct_query(struct nfct_handle *cth, int family, struct topologyd_private *priv)
{
        struct nfct_filter_dump *filter_dump = NULL;
        struct nf_conntrack *ct = NULL;

        ct = nfct_new(); //memory free is done at end of this function
	if (ct == NULL) {
		err("%s Failed to create nfct instance\n", __FUNCTION__);
		return;
	}

	nfct_set_attr_u8(ct, ATTR_ORIG_L3PROTO, family);
	if (family == AF_INET) {
		nfct_callback_register(cth, NFCT_T_ALL, host_node_stats_received, priv);
	} else {
		nfct_callback_register(cth, NFCT_T_ALL, host_node_ipv6_stats_received, priv);
	}

        filter_dump = nfct_filter_dump_create();
        if (filter_dump == NULL) {
		nfct_destroy(ct);
		err("%s Failed to allocate filter_dump", __FUNCTION__);
		return;
	}

        nfct_filter_dump_set_attr_u8(filter_dump,
                                     NFCT_FILTER_DUMP_L3NUM,
                                     family);

        nfct_query(cth, NFCT_Q_DUMP_FILTER, filter_dump);

        nfct_filter_dump_destroy(filter_dump);
	nfct_destroy(ct);
}

static void host_stat_ip_list_check_and_free(struct hlist_head *head, int node_to_delete)
{
	struct host_nf_tc_ip_list *expired_ip_entry = NULL;
	struct host_nf_tc_ip_list *n_ip = NULL;
	struct hlist_node *tmp = NULL;

	hlist_for_each_entry_safe(n_ip, tmp, head, hlist) {
		if (n_ip->in_use == node_to_delete) {
			expired_ip_entry = n_ip;
			hlist_del(&expired_ip_entry->hlist, head);
			free(expired_ip_entry);
		}
	}
}

void host_node_clean_ip_node_list(struct topologyd_private *priv, int ip_node_clean)
{
	int i;

	for (i = 0; i < CT_IP_NODE_HTABLE_SIZE ; i++) {
		if (!hlist_empty(&priv->host.ct_ip_node_htable[i])) {
			host_stat_ip_list_check_and_free(&priv->host.ct_ip_node_htable[i], ip_node_clean);
		}

		if (hlist_empty(&priv->host.ct_ipv6_node_htable[i]))
			continue;
		host_stat_ip_list_check_and_free(&priv->host.ct_ipv6_node_htable[i], ip_node_clean);
	}

}

static void host_node_clean_nf_tc(struct topologyd_private *priv, int ip_node_clean)
{
	int i;
	struct host_nf_tc *expired_entry = NULL;
	struct host_nf_tc *n = NULL;

	for (i = 0; i < CT_NODE_HTABLE_SIZE ; i++) {
		if ((hlist_empty(&priv->host.ct_node_htable[i])))
			continue;
		hlist_for_each_entry(n, &priv->host.ct_node_htable[i], hlist) {
			if(expired_entry) {
				hlist_del(&expired_entry->hlist, &priv->host.ct_node_htable[i]);
				host_node_ct_node_destroy(expired_entry);
				expired_entry = NULL;
			}
			if (!n->in_use)
				expired_entry = n;
		}
		if(expired_entry) {
			hlist_del(&expired_entry->hlist, &priv->host.ct_node_htable[i]);
			host_node_ct_node_destroy(expired_entry);
			expired_entry = NULL;
		}
	}

	host_node_clean_ip_node_list(priv, ip_node_clean);
}

int host_node_get_statistics(struct topologyd_private *priv)
{
        struct nfct_handle *cth;
	struct host_nf_tc *n = NULL;
	int i = 0;

	for (i = 0; i < CT_NODE_HTABLE_SIZE ; i++) {
		if ((hlist_empty(&priv->host.ct_node_htable[i])))
			continue;

                hlist_for_each_entry(n, &priv->host.ct_node_htable[i], hlist) {
			n->in_use = 0;
		}
	}
	host_node_clean_ip_node_list(priv, HOST_WAN_STATS_IP_NODE_NOMATCH);

	cth = nfct_open(CONNTRACK, 0);
        if (!cth) {
                err("Can't open handler\n");
		return -1;
	}

	host_node_nfct_query(cth, AF_INET, priv);
        nfct_close(cth);

	cth = nfct_open(CONNTRACK, 0);
        if (!cth) {
                err("Can't open handler\n");
		return -1;
	}
	host_node_nfct_query(cth, AF_INET6, priv);

        nfct_close(cth);
	host_node_clean_nf_tc(priv, HOST_WAN_STATS_IP_NODE_EXPIRED);
	return 0;
}
